#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import csv
import glob
import hashlib
import os
import re
import shutil
from datetime import datetime
from typing import Literal, Set, Tuple

import dask.dataframe as dask_df
import jsonlines
import pandas
from pydantic import ConfigDict, Field, computed_field, model_validator

from cliboa.adapter.csv import Csv
from cliboa.adapter.file import File
from cliboa.adapter.sqlite import SqliteAdapter
from cliboa.scenario.transform.file import FileBaseTransform
from cliboa.scenario.validator import EssentialParameters
from cliboa.util.base import _BaseObject  # _warn_deprecated_args
from cliboa.util.exception import FileNotFound, InvalidCount, InvalidParameter
from cliboa.util.string import StringUtil


class CsvColumnHash(FileBaseTransform):
    """
    Hash(SHA256) specific columns from csv file.
    """

    class Arguments(FileBaseTransform.Arguments):
        columns: list[str]

    def _stringToHash(self, string):
        return hashlib.sha256(string.encode()).hexdigest()

    def execute(self, *args):
        files = self.get_src_files()
        self.check_file_existence(files)

        self.io_files(files, func=self.convert)

    def convert(self, fi, fo):
        chunk_size_handling(self._read_csv_func, fi, fo)

    def _read_csv_func(self, chunksize, fi, fo):
        # Used in chunk_size_handling
        first_write = True
        tfr = pandas.read_csv(
            fi,
            dtype=str,
            encoding=self.args.encoding,
            chunksize=chunksize,
            na_filter=False,
        )
        for df in tfr:
            for c in self.args.columns:
                df[c] = df[c].apply(self._stringToHash)
            df.to_csv(
                fo,
                encoding=self.args.encoding,
                header=True if first_write else False,
                index=False,
                mode="w" if first_write else "a",
            )
            first_write = False


class CsvColumnExtract(FileBaseTransform):
    """
    Remove specific columns from csv file.
    """

    class Arguments(FileBaseTransform.Arguments):
        model_config = ConfigDict(coerce_numbers_to_str=True)

        columns: list[str] | None = None
        column_numbers: str | None = None

    def execute(self, *args):
        if not self.args.columns and not self.args.column_numbers:
            raise InvalidParameter("Specifying either 'column' or 'column_numbers' is essential.")
        if self.args.columns and self.args.column_numbers:
            raise InvalidParameter("Cannot specify both 'column' and 'column_numbers'.")

        files = self.get_src_files()
        self.check_file_existence(files)

        super().io_files(files, func=self.convert)

    def convert(self, fi, fo):
        if self.args.columns:
            Csv.extract_columns_with_names(fi, fo, self.args.columns)
        elif self.args.column_numbers:
            if isinstance(self.args.column_numbers, int) is True:
                remain_column_numbers = [self.args.column_numbers]
            else:
                column_numbers = self.args.column_numbers.split(",")
                remain_column_numbers = [int(n) for n in column_numbers]
            Csv.extract_columns_with_numbers(fi, fo, remain_column_numbers)


class CsvColumnDelete(FileBaseTransform):
    """
    Delete specific columns from csv file.
    """

    class Arguments(FileBaseTransform.Arguments):
        regex_pattern: str

    def execute(self, *args):
        files = self.get_src_files()
        self.check_file_existence(files)

        super().io_files(files, func=self.convert)

    def convert(self, fi, fo):
        chunk_size_handling(self._read_csv_func, fi, fo)

    def _read_csv_func(self, chunksize, fi, fo):
        # Used in chunk_size_handling
        first_write = True
        tfr = pandas.read_csv(
            fi,
            dtype=str,
            encoding=self.args.encoding,
            chunksize=chunksize,
            na_filter=False,
        )
        pattern = re.compile(self.args.regex_pattern)
        for df in tfr:
            for column in df.columns.values:
                if pattern.fullmatch(column):
                    df = df.drop(column, axis=1)
                # output an empty file when all columns are deleted
                if df.empty:
                    df = df.dropna(how="all")
            df.to_csv(
                fo,
                encoding=self.args.encoding,
                header=True if first_write else False,
                index=False,
                mode="w" if first_write else "a",
            )
            first_write = False


class CsvValueExtract(FileBaseTransform):
    """
    Extract a specific column from a CSV file and then replace it with a regular expression.
    """

    class Arguments(FileBaseTransform.Arguments):
        column_regex_pattern: dict[str, str]

    def execute(self, *args):
        files = self.get_src_files()
        self.check_file_existence(files)
        super().io_files(files, func=self.convert)

    def convert(self, fi, fo):
        with open(fi, mode="rt") as i:
            reader = csv.DictReader(i)
            with open(fo, mode="wt") as o:
                writer = csv.DictWriter(o, reader.fieldnames)
                writer.writeheader()
                for line in reader:
                    for column, regex_pattern in self.args.column_regex_pattern.items():
                        if re.search(regex_pattern, line[column]):
                            line[column] = re.search(regex_pattern, line[column]).group()
                        else:
                            line[column] = ""
                    writer.writerow(dict(line))


class CsvColumnConcat(FileBaseTransform):
    """
    Concat specific columns from csv file.
    """

    class Arguments(FileBaseTransform.Arguments):
        columns: list[str] = Field(min_length=2)
        dest_column_name: str
        sep: str = ""

    def execute(self, *args):
        files = self.get_src_files()
        self.check_file_existence(files)

        self.io_files(files, func=self.convert)

    def convert(self, fi, fo):
        chunk_size_handling(self._read_csv_func, fi, fo)

    def _read_csv_func(self, chunksize, fi, fo):
        # Used in chunk_size_handling
        first_write = True
        tfr = pandas.read_csv(
            fi,
            dtype=str,
            encoding=self.args.encoding,
            chunksize=chunksize,
            na_filter=False,
        )
        dest_str = None
        for df in tfr:
            for c in self.args.columns:
                if dest_str is None:
                    dest_str = df[c].astype(str)
                else:
                    dest_str = dest_str + self.args.sep + df[c].astype(str)
                df = df.drop(columns=[c])
            df[self.args.dest_column_name] = dest_str
            df.to_csv(
                fo,
                encoding=self.args.encoding,
                header=True if first_write else False,
                index=False,
                mode="w" if first_write else "a",
            )
            first_write = False


class CsvTypeConvert(FileBaseTransform):
    """
    Convert the type of specific column in a csv file.
    """

    class Arguments(FileBaseTransform.Arguments):
        dest_dir: str
        column: list[str]
        type: str

    def execute(self, *args):
        self.args.resolve_dest_dir()

        files = self.get_src_files()
        self.check_file_existence(files)
        self.logger.info("Files found %s" % files)
        self.io_files(files, func=self.convert)

    def convert(self, fi, fo):
        chunk_size_handling(self._read_csv_func, fi, fo)

    def _read_csv_func(self, chunksize, fi, fo):
        first_write = True
        tfr = pandas.read_csv(
            fi,
            dtype=object,
            encoding=self.args.encoding,
            chunksize=chunksize,
            na_filter=False,
        )
        for df in tfr:
            for column in self.args.column:
                try:
                    if self.args.type == "int":
                        # When reading from csv, the following error occurs:
                        # ValueError: invalid literal for int() with base 10
                        # To avoid this, convert to float and then convert to int
                        df[column] = df[column].astype("float")
                        df[column] = df[column].astype("int")
                    else:
                        df[column] = df[column].astype(self.args.type)
                except Exception:
                    raise InvalidParameter(
                        "Conversion to this type is not possible. %s" % self.args.type
                    )
            df.to_csv(
                fo,
                encoding=self.args.encoding,
                header=True if first_write else False,
                index=False,
                mode="w" if first_write else "a",
            )
            first_write = False


class CsvMergeExclusive(FileBaseTransform):
    """
    Compare specific columns each file.
    If matched, exclude rows.
    """

    def __init__(self, **kwargs):
        super().__init__(*kwargs)
        self.df_target_list = None

    class Arguments(FileBaseTransform.Arguments):
        target_compare_path: str
        src_column: str | None = None
        target_column: str | None = None
        all_column: bool = False

    def execute(self, *args):
        files = self.get_src_files()
        self.check_file_existence(files)

        target = self._resolve("adapter_file", File).get_target_files(
            os.path.dirname(self.args.target_compare_path),
            os.path.basename(self.args.target_compare_path),
        )
        self.check_file_existence(target)

        if self.args.all_column and (self.args.src_column or self.args.target_column):
            raise KeyError("all_column cannot coexist with src_column or target_column.")

        header = pandas.read_csv(self.args.target_compare_path, nrows=0)
        if self.args.all_column is False and self.args.target_column not in header:
            raise KeyError(
                "Target Compare file does not exist target column [%s]." % self.args.target_column
            )

        if self.args.all_column:
            df_target = pandas.read_csv(self.args.target_compare_path, dtype=str)
            self.df_target_list = df_target.values.tolist()
        else:
            df_target = pandas.read_csv(
                self.args.target_compare_path, usecols=[self.args.target_column], dtype=str
            )
            self.df_target_list = df_target[self.args.target_column].values.tolist()

        self.io_files(files, func=self.convert)

    def convert(self, fi, fo):
        if self.args.all_column is False:
            header = pandas.read_csv(fi, dtype=str, encoding=self.args.encoding, nrows=0)
            try:
                header[self.args.src_column].values.tolist()
            except KeyError:
                raise KeyError(
                    "Src file does not exist target column [%s]." % self.args.target_column
                )

        self._csv_write(fi, fo)

    def _csv_write(self, fi, fo):
        # Used in chunk_size_handling
        df = pandas.read_csv(fi, dtype=str, na_filter=False)
        if self.args.all_column:
            df_target_set = {hash(tuple(row)) for row in self.df_target_list}
            df = df.drop(self._all_elements_match(df.values.tolist(), df_target_set))
        else:
            df = df[~df[self.args.src_column].isin(self.df_target_list)]
        df.to_csv(
            fo,
            encoding=self.args.encoding,
            header=True,
            index=False,
            mode="a",
        )

    def _all_elements_match(self, df_src_list, df_target_set):
        return [i for i, row in enumerate(df_src_list) if hash(tuple(row)) in df_target_set]


class ColumnLengthAdjust(FileBaseTransform):
    """
    Adjust csv (tsv) column to maximum length
    """

    class Arguments(FileBaseTransform.Arguments):
        adjust: dict[str, int]

    def execute(self, *args):
        files = self.get_src_files()
        self.check_file_existence(files)

        self.io_writers(files, encoding=self.args.encoding)

        for fi, fo in super().io_writers(files, encoding=self.args.encoding):
            reader = csv.DictReader(fi)
            writer = csv.DictWriter(fo, reader.fieldnames)
            writer.writeheader()

            for row in reader:
                for k, v in self.args.adjust.items():
                    f1 = row.get(k)
                    if len(f1) > v:
                        row[k] = f1[:v]
                writer.writerow(row)
            fo.flush()


class CsvMerge(FileBaseTransform):
    """
    Merge two csv files
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._target_df = None

    class Arguments(FileBaseTransform.Arguments):
        dest_dir: str
        target_path: str
        join_on: str
        engine: Literal["pandas", "dask"] = "pandas"

    def execute(self, *args):
        self.args.resolve_dest_dir()

        source_files = self.get_src_files()
        if not self.check_file_existence(source_files):
            raise InvalidCount(
                "An input file %s does not exist."
                % os.path.join(self.args.src_dir, self.args.src_pattern)
            )

        target_files = glob.glob(self.args.target_path)
        if len(target_files) == 0:
            raise InvalidCount("An target file %s does not exist." % self.target_path)
        elif len(target_files) > 1:
            self.logger.error("Hit target files %s" % target_files)
            raise InvalidCount("Target files must be only one.")

        if self.args.engine == "dask":
            self._dask_merge(source_files, target_files[0])
        else:
            self._pandas_merge(source_files, target_files[0])

    def _dask_merge(self, source_files: list[str], target_file: str) -> None:
        """
        merge using dask
        """
        dd_target = dask_df.read_csv(target_file, encoding=self.args.encoding)

        for source_file in source_files:
            dest_path = os.path.join(self.args.dest_dir, os.path.basename(source_file))
            dd_source = dask_df.read_csv(source_file, encoding=self.args.encoding)
            merged_dd = dask_df.merge(dd_source, dd_target, on=self.args.join_on)
            merged_dd.to_csv(dest_path, single_file=True, index=False, encoding=self.args.encoding)

    def _pandas_merge(self, source_files: list[str], target_file: str) -> None:
        """
        merge using pandas
        NOTE: if target file is too large, use dask engine.
        """
        self._target_df = pandas.read_csv(target_file)
        for source_file in source_files:
            chunk_size_handling(self._pandas_merge_one, source_file)

    def _pandas_merge_one(self, chunksize: int, source_file: str) -> None:
        dest_path = os.path.join(self.args.dest_dir, os.path.basename(source_file))

        chunks = pandas.read_csv(source_file, chunksize=chunksize, encoding=self.args.encoding)
        is_first_chunk = True
        for chunk in chunks:
            merged_chunk = pandas.merge(chunk, self._target_df, on=self.args.join_on)

            if is_first_chunk:
                merged_chunk.to_csv(
                    dest_path, index=False, mode="w", header=True, encoding=self.args.encoding
                )
                is_first_chunk = False
            else:
                merged_chunk.to_csv(
                    dest_path, index=False, mode="a", header=False, encoding=self.args.encoding
                )


class CsvColumnSelect(FileBaseTransform):
    """
    Select columns in Csv file in specified order
    """

    class Arguments(FileBaseTransform.Arguments):
        column_order: list[str]

    def execute(self, *args):
        files = self.get_src_files()
        if not self.check_file_existence(files):
            raise FileNotFound("No files are found.")

        super().io_files(files, func=self.convert)

    def convert(self, fi, fo):
        chunk_size_handling(self._read_csv_func, fi, fo)

    def _read_csv_func(self, chunksize, fi, fo):
        # Used in chunk_size_handling
        first_write = True
        tfr = pandas.read_csv(
            fi,
            dtype=str,
            encoding=self.args.encoding,
            chunksize=chunksize,
            na_filter=False,
        )
        for df in tfr:
            if set(self.args.column_order) - set(df.columns.values):
                raise InvalidParameter(
                    "column_order define not included target file's column : %s"
                    % (set(self.args.column_order) - set(df.columns.values))
                )
            df = df.loc[:, self.args.column_order]
            df.to_csv(
                fo,
                encoding=self.args.encoding,
                header=True if first_write else False,
                index=False,
                mode="w" if first_write else "a",
            )
            first_write = False


class CsvConcat(FileBaseTransform):
    """
    Concat csv files
    """

    class Arguments(FileBaseTransform.Arguments):
        src_pattern: str | None = None
        src_filenames: list[str] | None = None
        dest_dir: str
        dest_name: str

        @model_validator(mode="before")
        def check_exclusive_fields(cls, data: dict) -> dict:
            if not isinstance(data, dict):
                raise ValueError(f"arguments is not dict: {data}")

            filenames_present = "src_filenames" in data
            pattern_present = "src_pattern" in data
            if not pattern_present and not filenames_present:
                raise InvalidParameter(
                    "Specifying either 'src_pattern' or 'src_filenames' is essential."
                )
            elif pattern_present and filenames_present:
                raise InvalidParameter("Cannot specify both 'src_pattern' and 'src_filenames'.")
            return data

    def execute(self, *args):
        if self.args.src_pattern:
            files = self.get_src_files()
        else:
            files = []
            for file in self.args.src_filenames:
                files.append(os.path.join(self.args.src_dir, file))

        if len(files) == 0:
            raise FileNotFound("No files are found.")
        elif len(files) == 1:
            self.logger.warning("Two or more input files are required.")

        # Create output headers to conform to the concat specification.
        file_1 = files[0]
        output_header = pandas.read_csv(
            file_1, dtype=str, encoding=self.args.encoding, nrows=0, na_filter=False
        )
        for file in files[1:]:
            output_header = pandas.concat(
                [
                    output_header,
                    pandas.read_csv(
                        file, dtype=str, encoding=self.args.encoding, nrows=0, na_filter=False
                    ),
                ]
            )

        chunk_size_handling(self._read_csv_func, files, output_header)

    def _read_csv_func(self, chunksize, files, output_header):
        # Used in chunk_size_handling
        first_write = True
        for file in files:
            tfr = pandas.read_csv(
                file,
                dtype=str,
                encoding=self.args.encoding,
                chunksize=chunksize,
                na_filter=False,
            )
            for df in tfr:
                # Change the header order to the one you plan to output.
                df = pandas.concat([output_header, df])
                df.to_csv(
                    os.path.join(self.args.resolve_dest_dir(), self.args.dest_name),
                    encoding=self.args.encoding,
                    header=True if first_write else False,
                    index=False,
                    mode="w" if first_write else "a",
                )
                first_write = False


class CsvConvert(FileBaseTransform):
    """
    Change csv format
    """

    class Arguments(FileBaseTransform.Arguments):
        headers: list[dict[str, str]] = Field(default_factory=list)
        headers_existence: bool = True
        add_headers: list[str] | None = None
        before_format: str = "csv"
        before_enc: str = "utf-8"
        before_escapechar: str | None = None
        after_format: str | None = None
        after_enc: str | None = None
        after_nl: str = "LF"
        after_escapechar: str | None = None
        reader_quote: str = "QUOTE_MINIMAL"
        quote: str = "QUOTE_MINIMAL"

    def execute(self, *args):
        if self.args.after_format is None:
            self.args.after_format = self.args.before_format

        if self.args.after_enc is None:
            self.args.after_enc = self.args.before_enc

        files = self.get_src_files()
        self.check_file_existence(files)

        self.io_files(files, ext=self.args.after_format, func=self.convert)

    def convert(self, fi, fo):
        with open(fi, mode="rt", encoding=self.args.before_enc) as i:
            reader = csv.reader(
                i,
                delimiter=Csv.delimiter_convert(self.args.before_format),
                quoting=Csv.quote_convert(self.args.reader_quote),
                escapechar=self.args.before_escapechar,
                doublequote=False if self.args.before_escapechar else True,
            )
            with open(fo, mode="wt", newline="", encoding=self.args.after_enc) as o:
                writer = csv.writer(
                    o,
                    delimiter=Csv.delimiter_convert(self.args.after_format),
                    quoting=Csv.quote_convert(self.args.quote),
                    lineterminator=Csv.newline_convert(self.args.after_nl),
                    escapechar=self.args.after_escapechar,
                    doublequote=False if self.args.after_escapechar else True,
                )

                for i, line in enumerate(reader):
                    if i == 0:
                        if self.args.headers_existence is False:
                            continue
                        if self.args.add_headers:
                            writer.writerow(self.args.add_headers)
                            writer.writerow(line)
                        else:
                            writer.writerow(self._replace_headers(line))
                    else:
                        writer.writerow(line)

    def _replace_headers(self, old_headers):
        """
        Replace old headers to new headers
        """
        if not self.args.headers:
            return old_headers

        converter = {}
        for headers in self.args.headers:
            for k, v in headers.items():
                converter[k] = v

        new_headers = []
        for oh in old_headers:
            r = converter.get(oh)
            new_headers.append(r if r is not None else oh)

        return new_headers


class CsvSort(FileBaseTransform):
    """
    Sort csv.
    """

    class Arguments(FileBaseTransform.Arguments):
        dest_dir: str
        order: list = []
        quote: str = "QUOTE_MINIMAL"
        no_duplicate: bool = False

    def execute(self, *args):
        self.args.resolve_dest_dir()

        files = self.get_src_files()
        self.check_file_existence(files)

        ymd_hms = datetime.now().strftime("%Y%m%d%H%M%S%f")
        dbname = ".%s_%s.db" % (ymd_hms, StringUtil().random_str(8))
        tblname = "temp_table"

        sqlite = SqliteAdapter()
        sqlite.connect(dbname)
        try:
            for file in files:
                _, filename = os.path.split(file)
                dest_file = os.path.join(self.args.dest_dir, filename)

                sqlite.import_table(file, tblname, encoding=self.args.encoding)
                sqlite.export_table(
                    tblname,
                    dest_file,
                    quoting=Csv.quote_convert(self.args.quote),
                    encoding=self.args.encoding,
                    order=self.args.order,
                    no_duplicate=self.args.no_duplicate,
                )
            sqlite.close()
        finally:
            os.remove(dbname)


class CsvToJsonl(FileBaseTransform):
    """
    Transform csv to jsonlines.
    """

    class Arguments(FileBaseTransform.Arguments):
        pass

    def execute(self, *args):
        files = self.get_src_files()
        self.check_file_existence(files)

        super().io_files(files, ext="jsonl", func=self.convert)

    def convert(self, fi, fo):
        with (
            open(fi, mode="r", encoding=self.args.encoding, newline="") as i,
            jsonlines.open(fo, mode="w") as writer,
        ):
            reader = csv.DictReader(i)
            for row in reader:
                writer.write(row)


class CsvColumnCopy(FileBaseTransform):
    """
    Copy column data (new or overwrite)
    """

    class Arguments(FileBaseTransform.Arguments):
        dest_dir: str
        src_column: str
        dest_column: str

    def execute(self, *args):
        files = self.get_src_files()
        self.check_file_existence(files)

        self.io_files(files, func=self.convert)

    def convert(self, fi, fo):
        header = pandas.read_csv(
            fi,
            dtype=str,
            encoding=self.args.encoding,
            nrows=0,
            na_filter=False,
        )
        if self.args.src_column not in header:
            raise KeyError("Copy source column does not exist in file. [%s]" % self.args.src_column)

        chunk_size_handling(self._read_csv_func, fi, fo)

    def _read_csv_func(self, chunksize, fi, fo):
        # Used in chunk_size_handling
        first_write = True
        tfr = pandas.read_csv(
            fi,
            dtype=str,
            encoding=self.args.encoding,
            chunksize=chunksize,
            na_filter=False,
        )

        for df in tfr:
            df[self.args.dest_column] = df[self.args.src_column]
            df.to_csv(
                fo,
                encoding=self.args.encoding,
                header=True if first_write else False,
                index=False,
                mode="w" if first_write else "a",
            )
            first_write = False


class CsvColumnReplace(FileBaseTransform):
    """
    Replace matching regular expression values for a specific column from a csv file.
    """

    class Arguments(FileBaseTransform.Arguments):
        column: str
        regex_pattern: str
        rep_str: str

        @computed_field
        @property
        def regex_compile(self) -> re.Pattern:
            return re.compile(self.regex_pattern)

    def _replace_string(self, string):
        return self.args.regex_compile.sub(self.args.rep_str, string)

    def execute(self, *args):
        files = self.get_src_files()
        self.check_file_existence(files)

        self.io_files(files, func=self.convert)

    def convert(self, fi, fo):
        header = pandas.read_csv(
            fi, dtype=str, encoding=self.args.encoding, nrows=0, na_filter=False
        )
        if self.args.column not in header:
            raise KeyError("Replace source column does not exist in file. [%s]" % self.args.column)

        chunk_size_handling(self._read_csv_func, fi, fo)

    def _read_csv_func(self, chunksize, fi, fo):
        # Used in chunk_size_handling
        first_write = True
        tfr = pandas.read_csv(
            fi,
            dtype=str,
            encoding=self.args.encoding,
            chunksize=chunksize,
            na_filter=False,
        )

        for df in tfr:
            df[self.args.column] = df[self.args.column].apply(self._replace_string)
            df.to_csv(
                fo,
                header=True if first_write else False,
                encoding=self.args.encoding,
                index=False,
                mode="w" if first_write else "a",
            )
            first_write = False


class CsvDuplicateRowDelete(FileBaseTransform):

    class Arguments(FileBaseTransform.Arguments):
        delimiter: str = ","
        engine: Literal["pandas", "dask"] = "pandas"

    def execute(self, *args):
        files = self.get_src_files()
        self.check_file_existence(files)

        self.io_files(files, func=self.convert)

    def convert(self, fi, fo):
        if self.args.engine == "dask":
            self._convert_with_dask(fi, fo)
        else:
            self._convert_with_pandas(fi, fo)

    def _convert_with_pandas(self, fi, fo):
        """
        Memory-efficient duplicate removal using pandas with set for tracking
        seen rows.

        Preserves original row order.
        """
        # Use existing chunk_size_handling infrastructure
        chunk_size_handling(self._read_csv_func, fi, fo)

    def _read_csv_func(self, chunksize, fi, fo):
        """
        Process CSV in chunks with duplicate removal.
        Used by chunk_size_handling for memory-efficient processing.
        """
        # Initialize state for each execution
        seen_rows = set()
        first_write = True

        for chunk in pandas.read_csv(
            fi,
            delimiter=self.args.delimiter,
            dtype=str,
            na_filter=False,
            chunksize=chunksize,
            header=None,
            encoding=self.args.encoding,
        ):
            # Convert DataFrame to list of tuples for hashability
            unique_rows = []
            for _, row in chunk.iterrows():
                row_tuple = tuple(row.values)
                if row_tuple not in seen_rows:
                    seen_rows.add(row_tuple)
                    unique_rows.append(row.values)

            # Write unique rows to output file
            if unique_rows:
                unique_df = pandas.DataFrame(unique_rows)
                unique_df.to_csv(
                    fo,
                    mode="w" if first_write else "a",
                    header=False,
                    index=False,
                    sep=self.args.delimiter,
                    encoding=self.args.encoding,
                )
                first_write = False

    def _convert_with_dask(self, fi, fo):
        """
        Handle very large files using dask.

        May not preserve original row order but provides better memory
        efficiency for extremely large datasets.
        """
        self.logger.info("Using dask engine - original row order may not be preserved")

        # Read CSV with dask
        df = dask_df.read_csv(
            fi,
            delimiter=self.args.delimiter,
            dtype=str,
            na_filter=False,
            encoding=self.args.encoding,
            header=None,
        )

        # Remove duplicates using dask
        df_unique = df.drop_duplicates()

        # Write to output file
        df_unique.to_csv(
            fo,
            single_file=True,
            header=False,
            index=False,
            sep=self.args.delimiter,
            encoding=self.args.encoding,
        )


class CsvRowDelete(FileBaseTransform):
    class Arguments(FileBaseTransform.Arguments):
        alter_path: str
        src_key_column: str
        alter_key_column: str
        delimiter: str = ","
        has_match: bool = True

    def execute(self, *args):
        files = self.get_src_files()

        self.check_file_existence(files)
        self.io_files(files, func=self.convert)

    def convert(self, fi, fo):
        s = set()
        with open(self.args.alter_path, "r") as al:
            alt_reader = csv.DictReader(al, delimiter=self.args.delimiter)
            for alt_row in alt_reader:
                s.add(alt_row[self.args.alter_key_column])
        with open(fi, "r") as i:
            reader = csv.DictReader(i, delimiter=self.args.delimiter)
            with open(fo, "w", newline="") as o:
                writer = csv.DictWriter(
                    o, fieldnames=reader.fieldnames, delimiter=self.args.delimiter
                )
                writer.writeheader()
                for row in reader:
                    if self.args.has_match is True and row[self.args.src_key_column] not in s:
                        writer.writerow(row)
                    elif self.args.has_match is False and row[self.args.src_key_column] in s:
                        writer.writerow(row)


class CsvSplit(FileBaseTransform):
    """
    Split csv files, using the value of a specific column as the output filename.

    - The output filename is the value from the specified column, with a .csv extension appended.
    - If the value of the specified column is empty, that row should be ignored.
    - When multiple source files exist, the column definitions must all be the same.
    """

    class Arguments(FileBaseTransform.Arguments):
        method: Literal["rows", "grouped"]
        key_column: str | None = None
        rows: int | None = None
        suffix_format: str = ".{:02d}"

    def execute(self, *args) -> None:
        files = self.get_src_files()
        self.check_file_existence(files)

        if self.args.method == "rows":
            executeInstance = _CsvSplitMethodRows(self.args)
        elif self.args.method == "grouped":
            executeInstance = _CsvSplitMethodGrouped(self.args)
        else:
            raise NotImplementedError(
                f"Defined {self.args.method} is not implemented logic in execute."
            )
        executeInstance.execute(files)


class _CsvSplitMethodBase(_BaseObject):
    def __init__(self, args):
        super().__init__()
        self.args = args

    def execute(self, files: list[str]) -> None:
        raise NotImplementedError("Please implement execute method.")


class _CsvSplitMethodRows(_CsvSplitMethodBase):
    def execute(self, files: list[str]) -> None:
        valid2 = EssentialParameters(
            self.__class__.__name__,
            [
                self.args.rows,
            ],
        )
        valid2()
        for filepath in files:
            self._split_one(filepath)

    def _split_one(self, filepath: str) -> None:
        self._logger.info("Split {:s} per {:d} rows".format(filepath, self.args.rows))
        file_name, ext = os.path.splitext(os.path.basename(filepath))
        with open(filepath, "r", encoding=self.args.encoding, newline="") as f_in:
            reader = csv.reader(f_in)
            try:
                header = next(reader)
            except StopIteration:
                self._logger.error(f"Empty {filepath}")
                return

            file_index = 0
            rows_count = 0
            writer = None
            f_out = None
            output_filepath = None

            for line in reader:
                # new file per self.args.rows
                if rows_count % self.args.rows == 0:
                    # close file-pointer if exists
                    if f_out:
                        f_out.close()
                        self._logger.info(
                            f"Generated {output_filepath} with {self.args.rows} rows"
                            f" by read up to line {rows_count} of the original."
                        )
                    # open new file-pointer
                    suffix = self.args.suffix_format.format(file_index)
                    output_filepath = os.path.join(
                        self.args.resolve_dest_dir(), f"{file_name}{suffix}{ext}"
                    )
                    f_out = open(output_filepath, "w", encoding=self.args.encoding, newline="")
                    writer = csv.writer(f_out)
                    writer.writerow(header)
                    file_index += 1
                # write current line
                writer.writerow(line)
                rows_count += 1
            # close last file-pointer if exists
            if f_out:
                f_out.close()
                last_rows = rows_count % self.args.rows
                if last_rows == 0:
                    last_rows = self.args.rows
                self._logger.info(
                    f"Generated {output_filepath} with {last_rows} rows"
                    f" by read up to line {rows_count} of the original."
                )


class _CsvSplitMethodGrouped(_CsvSplitMethodBase):
    def execute(self, files: list[str]) -> None:
        valid2 = EssentialParameters(
            self.__class__.__name__,
            [
                self.args.key_column,
            ],
        )
        valid2()

        if len(files) >= 2:
            self._validate_headers(files)

        unique_keys, found_empty = chunk_size_handling(self._collect_unique_keys, files)
        if not unique_keys:
            raise ValueError(
                "No valid keys found in the specified column. No files will be created."
            )
        self._logger.info(f"Found {len(unique_keys)} unique key(s). Starting file split process.")

        if len(unique_keys) == 1:
            chunk_size_handling(self._process_single_key, files, unique_keys.pop(), found_empty)
        else:
            chunk_size_handling(self._process_multiple_keys, files, found_empty)

    def _validate_headers(self, files: list[str]) -> None:
        reference_header = pandas.read_csv(
            files[0], nrows=0, encoding=self.args.encoding
        ).columns.tolist()

        for file_path in files[1:]:
            current_header = pandas.read_csv(
                file_path, nrows=0, encoding=self.args.encoding
            ).columns.tolist()
            if current_header != reference_header:
                raise ValueError(
                    f"Header mismatch found. File '{files[0]}' header is {reference_header}, "
                    f"but file '{file_path}' header is {current_header}."
                )

    def _collect_unique_keys(self, chunksize: int, files: list[str]) -> Tuple[Set[str], bool]:
        unique_keys = set()
        found_empty = False
        for file_path in files:
            with pandas.read_csv(
                file_path,
                chunksize=chunksize,
                usecols=[self.args.key_column],
                keep_default_na=False,
                encoding=self.args.encoding,
            ) as reader:
                for chunk in reader:
                    cleaned_keys = chunk[self.args.key_column].astype(str).str.strip()
                    if not found_empty and (cleaned_keys == "").any():
                        found_empty = True
                    non_empty_keys = cleaned_keys[cleaned_keys != ""].unique()
                    unique_keys.update(non_empty_keys)
        return unique_keys, found_empty

    def _process_single_key(
        self, chunksize: int, files: list[str], key: str, found_empty: bool
    ) -> None:
        self._logger.info(f"Processing in single-key mode. Outputting to '{key}.csv'.")
        output_filename = f"{key}.csv"
        output_path = os.path.join(self.args.resolve_dest_dir(), output_filename)

        write_mode = "w"
        if found_empty is False:
            files = files.copy()
            first_file = files.pop()
            shutil.copy2(first_file, output_path)
            if len(files) == 0:
                # When single input file and single key and not found empty value, very fast.
                return
            write_mode = "a"

        is_first_write = True
        with open(output_path, write_mode, newline="", encoding=self.args.encoding) as f_out:
            for file_path in files:
                with pandas.read_csv(
                    file_path,
                    chunksize=chunksize,
                    keep_default_na=False,
                    encoding=self.args.encoding,
                ) as reader:
                    for chunk in reader:
                        if found_empty is False:
                            chunk.to_csv(
                                f_out, header=False, index=False, encoding=self.args.encoding
                            )
                        else:
                            filtered_chunk = chunk[chunk[self.args.key_column].astype(str) == key]
                            if not filtered_chunk.empty:
                                filtered_chunk.to_csv(
                                    f_out,
                                    header=is_first_write,
                                    index=False,
                                    encoding=self.args.encoding,
                                )
                                is_first_write = False

    def _process_multiple_keys(self, chunksize: int, files: list[str], found_empty: bool) -> None:
        self._logger.info("Processing in multi-key mode.")
        file_pointers = {}
        try:
            for file_path in files:
                with pandas.read_csv(
                    file_path,
                    chunksize=chunksize,
                    keep_default_na=False,
                    encoding=self.args.encoding,
                ) as reader:
                    for chunk in reader:
                        if found_empty:
                            chunk.dropna(subset=[self.args.key_column], inplace=True)
                            chunk[self.args.key_column] = (
                                chunk[self.args.key_column].astype(str).str.strip()
                            )
                            chunk = chunk[chunk[self.args.key_column] != ""]
                        if chunk.empty:
                            continue

                        for key, group_df in chunk.groupby(self.args.key_column):
                            output_filename = f"{key}.csv"
                            if output_filename not in file_pointers:
                                file_pointers[output_filename] = open(
                                    os.path.join(self.args.resolve_dest_dir(), output_filename),
                                    "w",
                                    newline="",
                                    encoding=self.args.encoding,
                                )
                                group_df.to_csv(
                                    file_pointers[output_filename], header=True, index=False
                                )
                            else:
                                group_df.to_csv(
                                    file_pointers[output_filename], header=False, index=False
                                )
        finally:
            for fp in file_pointers.values():
                fp.close()


def chunk_size_handling(read_csv_func, *args, **kwd):
    """
    Processing to avoid memory errors in pandas's read_csv.
    Use this function when you want to do the same handling when extending cliboa.
    """
    chunksize = 1024 * 1024
    while 0 < chunksize:
        try:
            return read_csv_func(chunksize, *args, **kwd)
        except MemoryError as error:
            if chunksize <= 1:
                raise error
            chunksize //= 2



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import copy
import inspect
from abc import abstractmethod
from typing import Any

from cliboa import state
from cliboa.core.interface import _IContext, _IExecute
from cliboa.core.model import CommandArgument, StepModel
from cliboa.listener.interface import IScenarioExecutor
from cliboa.scenario.base import BaseStep
from cliboa.scenario.interface import IParentStep
from cliboa.util.base import _BaseObject
from cliboa.util.cache import ObjectStore
from cliboa.util.constant import StepStatus


class _BaseExecutor(_BaseObject, _IExecute):
    """
    Base class of executor.
    Execute main logic with handling listeners.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._listeners = []

    def register_listener(self, listener) -> None:
        self._listeners.append(listener)

    def _get_listener_arg(self) -> Any:
        return self

    def _execute_listener(self, kind: str, e: Exception | None = None) -> None:
        for lis in self._listeners:
            try:
                target = getattr(lis, kind)
                if callable(target):
                    if e:
                        target(self._get_listener_arg(), e)
                    else:
                        target(self._get_listener_arg())
            except Exception:
                self._logger.exception(
                    f"Error occurred during {lis.__class__.__name__}.{kind}"
                    f" in {self.__class__.__name__}"
                )

    def execute(self) -> int | None:
        """
        Execute with listeners.
        """
        try:
            self._execute_listener("before")
            res = self._execute_main()
            self._execute_listener("after")
            return res
        except Exception as e:
            self._logger.exception(
                "{}({}) occurred during the execution of {}.{}".format(
                    e.__class__.__name__,
                    str(e),
                    self.__class__.__module__,
                    self.__class__.__name__,
                )
            )
            self._execute_listener("error", e)
            return StepStatus.ABNORMAL_TERMINATION
        finally:
            self._execute_listener("completion")

    @abstractmethod
    def _execute_main(self) -> int | None:
        raise NotImplementedError()


class _ScenarioExecutor(_BaseExecutor, IScenarioExecutor):
    """
    Executor for scenario
    """

    def __init__(self, steps: list[_IExecute], **kwargs):
        super().__init__(**kwargs)
        self._steps = steps
        self._max_steps_size = len(steps)

    @property
    def max_steps_size(self) -> int:
        """
        Get max inisteps size (set during initialization).
        """
        return self._max_steps_size

    @property
    def current_steps_size(self) -> int:
        """
        Get current remaining steps size.
        """
        return len(self._steps)

    def _execute_main(self) -> int:
        """
        Wrap steps with _StepExecutor and execute.
        """
        state.set_steps_max(self.current_steps_size)
        state.set_steps_current(0)
        state.set_in_steps(True)
        res = None
        while self.current_steps_size > 0:
            step = self._steps.pop(0)
            state.set_steps_current(state.steps_max - self.current_steps_size)
            res = step.execute()
            if res is None:
                continue
            elif res == StepStatus.SUCCESSFUL_TERMINATION:
                self._logger.info("Step response [successful termination]. Scenario will be end.")
                break
            else:
                self._logger.error(
                    f"Step response [abnormal termination: {res}]. Scenario will be end."
                )
                break
        state.set_in_steps(False)
        return StepStatus.SUCCESSFUL_TERMINATION if res is None else res


class _StepExecutor(_BaseExecutor, IParentStep):
    """
    Executor for step
    """

    def __init__(
        self,
        step: BaseStep,
        model: StepModel,
        cmd_arg: CommandArgument | None = None,
        context: _IContext | None = None,
        symbol_model: StepModel | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        step.parent = self
        step._set_arguments(model.arguments)
        self._step = step
        self._model = model
        self._context = context
        self._symbol_model = symbol_model
        self._exec_args = copy.deepcopy(cmd_arg.args) if cmd_arg and cmd_arg.args else []
        self._exec_kwargs = copy.deepcopy(cmd_arg.kwargs) if cmd_arg and cmd_arg.kwargs else {}

    @property
    def step(self) -> BaseStep:
        return self._step

    @property
    def step_name(self) -> str:
        return self._model.step

    @property
    def symbol_name(self) -> str | None:
        return self._model.symbol

    def get_symbol_arguments(self) -> dict[str, Any]:
        if self._symbol_model:
            return self._symbol_model.arguments
        else:
            return {}

    def put_to_context(self, value: Any) -> None:
        if self._context:
            self._context.put(self.step_name, value)
        # v2 backward compability
        ObjectStore.put(self.step_name, value, quiet=True)

    def get_from_context(self, target: str | None = None) -> Any:
        if target is None:
            target = self.symbol_name
        if self._context:
            return self._context.get(target)

    def _get_listener_arg(self) -> Any:
        return self.step

    def _execute_main(self) -> int | None:
        candidates = (
            (self._exec_args, self._exec_kwargs),
            ([], self._exec_kwargs),
            (self._exec_args, {}),
        )
        sig = inspect.signature(self.step.execute)
        for args, kwargs in candidates:
            try:
                sig.bind(*args, **kwargs)
            except TypeError:
                continue
            return self.step.execute(*args, **kwargs)
        return self.step.execute()



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
"""
Module directly under cliboa should only handle global states that are not deprecated.
Operationally, this layer is expected to be treated as virtually identical to the utils layer.
"""

from .util.state import state  # noqa



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import bz2
import csv
import gzip
import os
import re
import shutil
import tarfile
import tempfile
import zipfile
from typing import Literal

import pandas

from cliboa.adapter.file import File
from cliboa.scenario.file import FileRead, FileWrite
from cliboa.util.base import _warn_deprecated_args
from cliboa.util.date import DateUtil
from cliboa.util.exception import CliboaException, InvalidParameter


class FileBaseTransform(FileRead, FileWrite):
    """
    Base class of file transform classes

    Basically transform class of Cliboa is that find files,
    do something, and output transformed files.
    When output files are created, name of the files will be the same name with the input files
    (original input files will be changed to the transformed files).
    If you would not like to remove the original files,
    give a path to the "dest_dir" for output directory.
    (If a non-existent directory path is specified, the directory is automatically created.)

    Note:
    Output files are not always be the same name with the input file names.
    See documentation for individual classes for details.
    """

    class Arguments(FileRead.Arguments, FileWrite.Arguments):
        force_continue: bool = False

        def resolve_dest_dir(self) -> str:
            if self.dest_dir:
                os.makedirs(self.dest_dir, exist_ok=True)
                return self.dest_dir
            else:
                return self.src_dir

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _force_continue(self):
        return self.args.force_continue

    def execute(self):
        pass

    def check_output_path(self, input_path, ext):
        root, name = os.path.split(input_path)

        if ext:
            if ext.startswith("."):
                output_name = os.path.splitext(name)[0] + ext
            else:
                output_name = os.path.splitext(name)[0] + "." + ext
        else:
            output_name = name

        if self.args.dest_dir:
            os.makedirs(self.args.dest_dir, exist_ok=True)
            output_dir = self.args.dest_dir
        else:
            output_dir = root

        output_path = os.path.join(output_dir, output_name)

        fd, temp_file = tempfile.mkstemp()
        os.close(fd)
        return output_path, temp_file

    def overwrite_output_path(self, input_path, output_path, temp_file):
        if input_path == output_path:
            os.remove(input_path)
        shutil.move(temp_file, output_path)

    def io_files(self, iterable, ext=None, func=None):
        """
        Iteration which returns input and output path.
        If the parameter "dest_dir" was given, the output file will be created under
        the given directory and returns input and output path.
        If not, the output file will be created to the same directory to the input file.

        Arguments:
            iterable (list): Input file list
            ext=None (str): Set an extension for output file,
                            if input and output extension would like to be changed.
                            "." is not necessary.
        """
        for input_path in iterable:
            output_path, temp_file = self.check_output_path(input_path, ext)

            try:
                func(input_path, temp_file)
            except Exception as e:
                if self.args.force_continue is True:
                    self.handle_error(e, input_path)
                else:
                    raise e

            self.overwrite_output_path(input_path, output_path, temp_file)

    def io_writers(self, iterable, mode="t", encoding="utf-8", ext=None):
        """
        Iteration which returns input and output writer.
        If the parameter "dest_dir" was given, the output file will be created under
        the given directory and returns input and output writer.
        If not, the output file will be created to the same directory to the input file.

        Arguments:
            iterable (list): Input file list
            mode="t" (str): Mode to open input files, opening file with text mode by default.
            encoding="utf-8" (str): File encoding. It will be ignored when mode is "b"
            ext=None (str): Set an extension for output file,
                            if input and output extension would like to be changed.
                            "." is not necessary.
        """
        if not "t" == mode and not "b" == mode:
            raise InvalidParameter("Unknown mode. One of the following is allowed [t, b]")

        if "b" in mode:
            encoding = None

        for input_path in iterable:
            output_path, temp_file = self.check_output_path(input_path, ext)

            if mode == "t":
                with (
                    open(input_path, mode="r", encoding=encoding, newline="") as i,
                    open(temp_file, mode="w", encoding=encoding, newline="") as o,
                ):
                    yield i, o
            elif mode == "b":
                with open(input_path, mode="rb") as i, open(temp_file, mode="wb") as o:
                    yield i, o

            self.overwrite_output_path(input_path, output_path, temp_file)

    def handle_error(self, e: Exception, input_path: str):
        # Please implement in a subclass if you would like to do something.
        self.logger.warning(
            (
                e,
                input_path,
            )
        )


class FileDecompress(FileBaseTransform):
    """
    Decompress the specified file
    """

    class Arguments(FileBaseTransform.Arguments):
        chunk_size: int | None = None
        password: str | None = None

    def execute(self, *args):
        files = self.get_src_files()
        self.check_file_existence(files)

        for f in files:
            _, ext = os.path.splitext(f)
            if ext == ".zip":
                self.logger.info("Decompress zip file %s" % f)
                if self.args.password is not None:
                    pwd = self.args.password.encode(self.args.encoding)
                else:
                    pwd = self.args.password
                with zipfile.ZipFile(f) as zp:
                    zp.extractall(self.args.resolve_dest_dir(), pwd=pwd),  # nosec
            elif ext == ".tar":
                self.logger.info("Decompress tar file %s" % f)
                with tarfile.open(f, "r:*") as tf:
                    tf.extractall(self.args.resolve_dest_dir())  # nosec
            elif ext == ".bz2":
                self.logger.info("Decompress bz2 file %s" % f)
                dcom_name = os.path.splitext(os.path.basename(f))[0]
                decom_path = os.path.join(self.args.resolve_dest_dir(), dcom_name)
                with bz2.open(f, mode="rb") as i, open(decom_path, mode="wb") as o:
                    while True:
                        buf = i.read(self.args.chunk_size)
                        if buf == b"":
                            break
                        o.write(buf)
            elif ext == ".gz":
                self.logger.info("Decompress gz file %s" % f)
                dcom_name = os.path.splitext(os.path.basename(f))[0]
                decom_path = os.path.join(self.args.resolve_dest_dir(), dcom_name)
                with gzip.open(f, "rb") as i, open(decom_path, "wb") as o:
                    while True:
                        buf = i.read(self.args.chunk_size)
                        if buf == b"":
                            break
                        o.write(buf)
            else:
                raise CliboaException("Unmatched any available decompress type %s" % f)


class FileCompress(FileBaseTransform):
    """
    Compress files
    """

    class Arguments(FileBaseTransform.Arguments):
        format: Literal["zip", "gz", "gzip", "bz2", "bzip2"]
        chunk_size: int = 1048576

    def execute(self, *args):
        files = self.get_src_files()
        self.check_file_existence(files)

        dest_dir = self.args.resolve_dest_dir()
        for f in files:
            if self.args.format == "zip":
                self.logger.info("Compress file %s to zip." % f)
                with zipfile.ZipFile(
                    os.path.join(dest_dir, (os.path.basename(f) + ".zip")),
                    "w",
                    zipfile.ZIP_DEFLATED,
                ) as o:
                    o.write(f, arcname=os.path.basename(f))
            elif self.args.format in ("gz", "gzip"):
                self.logger.info("Compress file %s to gzip." % f)
                com_path = os.path.join(dest_dir, (os.path.basename(f) + ".gz"))
                with open(f, "rb") as i, gzip.open(com_path, "wb") as o:
                    while True:
                        buf = i.read(self.args.chunk_size)
                        if buf == b"":
                            break
                        o.write(buf)
            elif self.args.format in ("bz2", "bzip2"):
                self.logger.info("Compress file %s to bzip2." % f)
                com_path = os.path.join(dest_dir, (os.path.basename(f) + ".bz2"))
                with open(f, "rb") as i, bz2.open(com_path, "wb") as o:
                    while True:
                        buf = i.read(self.args.chunk_size)
                        if buf == b"":
                            break
                        o.write(buf)


class DateFormatConvert(FileBaseTransform):
    """
    Convert csv (tsv) date field columns to another date field format columns
    """

    class Arguments(FileBaseTransform.Arguments):
        columns: list[str]
        formatter: str

    def execute(self, *args):
        files = self.get_src_files()
        self.check_file_existence(files)

        _, ext = os.path.splitext(files[0])
        if ext == ".csv":
            delimiter = ","
        elif ext == ".tsv":
            delimiter = "\t"

        for ins, ous in self.io_writers(files, encoding=self.args.encoding):
            reader = csv.DictReader(ins, delimiter=delimiter)
            writer = csv.DictWriter(ous, reader.fieldnames)
            writer.writeheader()
            date_util = DateUtil()
            for row in reader:
                for column in self.args.columns:
                    r = row.get(column)
                    if not r:
                        continue
                    row[column] = date_util.convert_date_format(r, self.args.formatter)
                writer.writerow(row)


class ExcelConvert(FileBaseTransform):
    """
    Convert excel to other format
    """

    class Arguments(FileBaseTransform.Arguments):
        pass

    def execute(self, *args):
        files = self.get_src_files()
        self.check_file_existence(files)

        # TODO Currently only excel to csv is supported.
        self.io_files(files, ext="csv", func=self.convert)

    def convert(self, fi, fo):
        self.logger.info("Convert %s to %s" % (fi, fo))
        df = pandas.read_excel(fi)
        df.to_csv(fo, encoding=self.args.encoding)


class FileCopy(FileBaseTransform):
    """
    Copy the file
    """

    class Arguments(FileBaseTransform.Arguments):
        dest_dir: str

    def execute(self, *args):
        files = self.get_src_files()
        self.check_file_existence(files)

        dest_dir = self.args.resolve_dest_dir()
        for file in files:
            shutil.copy(file, dest_dir)


class FileDivide(FileBaseTransform):
    """
    Divide a file to plural files
    """

    class Arguments(FileBaseTransform.Arguments):
        divide_rows: int
        header: bool = False
        suffix_pattern: str = ".%d"

    def divide_rows(self, divide_rows):
        self._divide_rows = divide_rows

    def header(self, header):
        self._header = header

    def suffix_pattern(self, suffix_pattern):
        self._suffix_pattern = suffix_pattern

    def execute(self, *args):
        files = self.get_src_files()
        self.check_file_existence(files)

        for file in files:
            fname = os.path.basename(file)

            px = ""
            if fname.startswith("."):
                fname = fname[1:]
                px = "."

            if "." in fname:
                nameonly, ext = fname.split(".", 1)
                ext = "." + ext
            else:
                nameonly = fname
                ext = ""

            if self.args.header:
                with open(file, encoding=self.args.encoding) as i:
                    self._header_row = i.readline()

            row = self._ifile_reader(file)
            newfilename = px + nameonly + self.args.suffix_pattern + ext

            dest_dir = self.args.resolve_dest_dir()
            has_left = True
            index = 1
            while has_left:
                ofile_path = os.path.join(dest_dir, newfilename % index)
                has_left = self._ofile_generator(ofile_path, row)
                index = index + 1

    def _ifile_reader(self, filepath):
        with open(filepath, encoding=self.args.encoding) as i:
            if self.args.header is True:
                i.readline()
            for line in i:
                yield line

    def _ofile_generator(self, filepath, row):
        left = False
        written = False
        with open(filepath, mode="w", encoding=self.args.encoding) as o:
            if self.args.header is True:
                o.write(self._header_row)
            for i, line in enumerate(row):
                written = True
                o.write(line)
                if i + 1 >= self.args.divide_rows:
                    left = True
                    break
        if written is False:
            os.remove(filepath)
        return left


class FileRename(FileBaseTransform):
    """
    Change file names with adding either prefix or suffix.
    """

    class Arguments(FileBaseTransform.Arguments):
        prefix: str = ""
        suffix: str = ""
        regex_pattern: str | None = None
        rep_str: str | None = None
        ext: str = ""
        without_ext: bool = True

    def prefix(self, prefix):
        self._prefix = prefix

    def suffix(self, suffix):
        self._suffix = suffix

    def regex_pattern(self, regex_pattern):
        self._regex_pattern = regex_pattern

    def rep_str(self, rep_str):
        self._rep_str = rep_str

    def ext(self, ext):
        self._ext = ext

    def without_ext(self, without_ext):
        self._without_ext = without_ext

    def execute(self, *args):
        files = self.get_src_files()
        self.check_file_existence(files)
        if self.args.rep_str is None and self.args.regex_pattern is not None:
            raise InvalidParameter("Replace string is not defined in yaml file: rep_str")
        if self.args.rep_str is not None and self.args.regex_pattern is None:
            raise InvalidParameter(
                "The conversion pattern is not defined in yaml file: regex_pattern"
            )

        for file in files:
            dirname = os.path.dirname(file)
            basename = os.path.basename(file)

            if self.args.without_ext is False:
                if self.args.prefix != "" or self.args.suffix != "" or self.args.ext != "":
                    raise InvalidParameter("Cannot be specified in without_ext mode")
                self._replace(
                    dirname,
                    file,
                    os.path.join(
                        dirname, re.sub(self.args.regex_pattern, self.args.rep_str, basename)
                    ),
                )
            else:
                px = ""
                if basename.startswith("."):
                    basename = basename[1:]
                    px = "."

                if "." in basename:
                    nameonly, extension = basename.split(".", 1)
                    extension = "." + extension
                else:
                    nameonly = basename
                    extension = ""

                if self.args.regex_pattern is not None and self.args.rep_str is not None:
                    nameonly = re.sub(self.args.regex_pattern, self.args.rep_str, nameonly)

                if self.args.ext:
                    extension = "." + self.args.ext

                newfilename = self.args.prefix + px + nameonly + self.args.suffix + extension
                self._replace(dirname, file, newfilename)

    def _replace(self, dirname, file, newfilename):
        newfilepath = os.path.join(dirname, newfilename)
        os.rename(file, newfilepath)
        self.logger.info("File name changed %s -> %s" % (file, newfilepath))


class FileConvert(FileBaseTransform):
    """
    Convert file encoding
    """

    class Arguments(FileBaseTransform.Arguments):
        encoding_from: str
        encoding_to: str
        errors: str | None = None

    def execute(self, *args):
        files = self.get_src_files()
        self.check_file_existence(files)

        self.io_files(files, func=self.convert)

    def convert(self, fi, fo):
        File().convert_encoding(
            fi,
            fo,
            self.args.encoding_from,
            self.args.encoding_to,
            self.args.errors,
        )

        self.logger.info("Encoded file %s" % fi)


class FileArchive(FileBaseTransform):
    """
    Create archeve object.a
    """

    class Arguments(FileBaseTransform.Arguments):
        dest_name: str
        format: Literal["zip", "tar"]
        create_dir: bool = False

    def execute(self, *args):
        files = self.get_src_files()
        self.check_file_existence(files)

        dest_dir = self.args.resolve_dest_dir()

        dest_path = os.path.join(dest_dir, (self.args.dest_name + ".%s" % self.args.format))

        if self.args.format == "tar":
            with tarfile.open(dest_path, "w") as tar:
                for file in files:
                    arcname = (
                        os.path.join(self.args.dest_name, os.path.basename(file))
                        if self.args.create_dir
                        else os.path.basename(file)
                    )
                    tar.add(file, arcname=arcname)
        elif self.args.format == "zip":
            with zipfile.ZipFile(dest_path, "w") as zp:
                for file in files:
                    arcname = (
                        os.path.join(self.args.dest_name, os.path.basename(file))
                        if self.args.create_dir
                        else os.path.basename(file)
                    )
                    zp.write(file, arcname=arcname)
        else:
            raise InvalidParameter("'format' must set one of the followings [tar, zip]")



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import json
import re

from pydantic import BaseModel

from cliboa import state
from cliboa.conf import env
from cliboa.listener.base import BaseStepListener
from cliboa.scenario.base import BaseStep


class StepStatusListener(BaseStepListener):
    """
    This listener is only for logging.
    By default, Cliboa implements StepStatusListener in all steps.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        mask = partial_mask = pattern = partial_pattern = None
        partial_num = 3
        try:
            mask = env.get("LOGGING_MASK", ".*password.*|.*secret_key.*")
            partial_mask = env.get("LOGGING_PARTIAL_MASK", ".*access_key.*|.*token.*")
            partial_num = int(env.get("LOGGING_PARTIAL_NUM", 3))
            pattern = re.compile(mask)
            partial_pattern = re.compile(partial_mask)
        except Exception as e:
            self.logger.warning(e)
        self._pattern = pattern
        self._partial_pattern = partial_pattern
        self._partial_num = partial_num

    def before(self, step: BaseStep) -> None:
        state.set(step.__class__.__name__)
        props_dict = {}
        props_values = step.__dict__.copy()
        if isinstance(props_values.get("_args"), BaseModel):
            args = props_values.pop("_args")
            props_values.update(args.model_dump())
        for k, v in props_values.items():
            if k in (
                "_di_map",
                "_di_kwargs",
                "_logger",
                "_parent",
                "_args",
                "_deprecated_warn_log",
            ):
                continue
            if v is not None and self._pattern is not None and self._pattern.search(k):
                props_dict[k] = "****"
            elif (
                v is not None
                and self._partial_pattern is not None
                and self._partial_pattern.search(k)
            ):
                orig_value = str(v)
                if len(orig_value) > self._partial_num * 2:
                    props_dict[k] = (
                        orig_value[: self._partial_num] + "****" + orig_value[-self._partial_num :]
                    )
                elif len(orig_value) > self._partial_num:
                    props_dict[k] = "****" + orig_value[-self._partial_num :]
                else:
                    props_dict[k] = "****"
            else:
                props_dict[k] = v
        self.logger.info(
            "Step properties: %s" % json.dumps(props_dict, ensure_ascii=False, default=str)
        )
        self.logger.info("Start step execution. %s" % step.__class__.__name__)

    def after(self, step: BaseStep) -> None:
        self.logger.info("Finish step execution. %s" % step.__class__.__name__)

    def completion(self, step: BaseStep) -> None:
        self.logger.info("Complete step execution. %s" % step.__class__.__name__)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import logging
from abc import abstractmethod

from cliboa.listener.interface import IScenarioExecutor
from cliboa.scenario.base import BaseStep
from cliboa.util.base import _BaseObject


class BaseListener(_BaseObject):
    """
    Base listener for all the listener classes
    """

    @property
    def logger(self) -> logging.Logger:
        return self._logger

    @abstractmethod
    def before(self, obj) -> None:
        """
        Execute before main logic start.
        """
        pass

    @abstractmethod
    def after(self, obj) -> None:
        """
        Execute after main logic end normally.
        """
        pass

    @abstractmethod
    def error(self, obj, e: Exception) -> None:
        """
        Execute after main logic raises Exception.
        """
        pass

    @abstractmethod
    def completion(self, obj) -> None:
        """
        Execute after main logic always.
        """
        pass


class BaseScenarioListener(BaseListener):
    """
    Listener for scenario
    """

    def before(self, executor: IScenarioExecutor) -> None:
        pass

    def after(self, executor: IScenarioExecutor) -> None:
        pass

    def error(self, executor: IScenarioExecutor, e: Exception) -> None:
        pass

    def completion(self, executor: IScenarioExecutor) -> None:
        pass


class BaseStepListener(BaseListener):
    """
    Listener for step
    """

    def before(self, step: BaseStep) -> None:
        pass

    def after(self, step: BaseStep) -> None:
        pass

    def error(self, step: BaseStep, e: Exception) -> None:
        pass

    def completion(self, step: BaseStep) -> None:
        pass



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from functools import partial

from cliboa.listener.base import BaseStepListener
from cliboa.scenario.base import BaseStep
from cliboa.util.base import _warn_deprecated

_warn_deprecated_step_listener = partial(
    _warn_deprecated,
    "cliboa.core.listener.StepListener",
    "3.0",
    "4.0",
    "cliboa.listener.base.BaseStepListener",
)


class StepListener(BaseStepListener):
    """
    for v2 backward compatible
    """

    def __init__(self, *args, **kwargs):
        _warn_deprecated_step_listener()
        super().__init__(*args, **kwargs)

    def before(self, step: BaseStep) -> None:
        if hasattr(self, "before_step"):
            _warn_deprecated_step_listener()
            self.before_step(step)

    def after(self, step: BaseStep) -> None:
        if hasattr(self, "after_step"):
            _warn_deprecated_step_listener()
            self.after_step(step)

    def error(self, step: BaseStep, e: Exception) -> None:
        if hasattr(self, "error_step"):
            _warn_deprecated_step_listener()
            self.error_step(step)

    def completion(self, step: BaseStep) -> None:
        if hasattr(self, "after_completion"):
            _warn_deprecated_step_listener()
            self.after_completion(step)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from pydantic import BaseModel

from cliboa.scenario.base import BaseStep
from cliboa.util.base import _warn_deprecated_args


class BaseAws(BaseStep):
    """
    Base class of AWS related classes
    """

    class Arguments(BaseModel):
        region: str
        access_key: str | None = None
        secret_key: str | None = None
        profile: str | None = None

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _region(self):
        return self.args.region

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _access_key(self):
        return self.args.access_key

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _secret_key(self):
        return self.args.secret_key

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _profile(self):
        return self.args.profile


class BaseS3(BaseAws):
    """
    Base class of S3 related classes
    """

    class Arguments(BaseAws.Arguments):
        region: str | None = None
        bucket: str
        role_arn: str | None = None
        external_id: str | None = None

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _bucket(self):
        return self.args.bucket

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _role_arn(self):
        return self.args.role_arn

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _external_id(self):
        return self.args.external_id



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import argparse
import logging
import logging.config
import os
from typing import Tuple

from cliboa.conf import env
from cliboa.core.manager import ScenarioManager
from cliboa.core.model import CommandArgument
from cliboa.util.base import _warn_deprecated, _warn_removed
from cliboa.util.exception import InvalidFormat
from cliboa.util.log import CliboaLogRecord


def _parse_args():
    """
    Parse command line arguments
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "project_name", default="hoge", help="Specify execution target project name"
    )
    parser.add_argument("execute_method_argument", nargs="*", default="")
    parser.add_argument(
        "-f",
        "--format",
        nargs="?",
        default="yaml",
        help="Specify yaml or json as FORMAT. Default foramt is yaml",
    )
    args = parser.parse_args()
    return args


def _initialize_cliboa_logging():
    """
    initialize cliboa logging - load logging.conf and set CliboaLogRecord.
    """
    config_dict = env.get("LOGGING_CONFIG_DICT")
    if config_dict:
        logging.config.dictConfig(config_dict)
    else:
        base_dir = env.get("BASE_DIR")
        config_path = env.get(
            "LOGGING_CONFIG_PATH",
            os.path.join(base_dir, "conf", "logging.conf") if base_dir is not None else None,
        )
        if config_path:
            logging.config.fileConfig(config_path, disable_existing_loggers=False)
    logging.setLogRecordFactory(CliboaLogRecord)


def _generate_scenario_path(project_name: str, scenario_format: str) -> Tuple[str, str]:
    """
    generate project's scenario path and common scenario path from project_name and scenario_format.
    """
    if scenario_format == "yaml":
        ext = env.get("SCENARIO_YAML_EXT", ".yml")
        pj_scenario_file = os.path.join(env.PROJECT_DIR, project_name, env.SCENARIO_FILE_NAME) + ext
        cmn_scenario_file = os.path.join(env.COMMON_DIR, env.SCENARIO_FILE_NAME) + ext
    elif scenario_format == "json":
        ext = "." + scenario_format
        pj_scenario_file = os.path.join(env.PROJECT_DIR, project_name, env.SCENARIO_FILE_NAME) + ext
        cmn_scenario_file = os.path.join(env.COMMON_DIR, env.SCENARIO_FILE_NAME) + ext
    else:
        raise InvalidFormat(f"scenario format '{scenario_format}' is invalid.")

    return pj_scenario_file, cmn_scenario_file


def _warn_backward_compatibility():
    if env.get("SCENARIO_DIR_NAME"):
        _warn_deprecated(
            "{CLIBOA_ENV}.SCENARIO_DIR_NAME",
            "3.0",
            "4.0",
            "{CLIBOA_ENV}.PROJECT_SCENARIO_DIR_NAME",
        )
    if env.get("SYSTEM_APPEND_PATHS"):
        _warn_removed("{CLIBOA_ENV}.SYSTEM_APPEND_PATHS", "3.0")
    base_dir = env.get("BASE_DIR")
    if base_dir and os.path.exists(os.path.join(base_dir, "conf", "cliboa.ini")):
        _warn_removed("configuration in cliboa.ini", "3.0")


def run():
    """
    The primary entry point function for cliboa.

    It's recommended to call this function if you want to use the cliboa application easily.
    """
    _warn_backward_compatibility()
    _initialize_cliboa_logging()
    cmd_args = _parse_args()

    pj, cmn = _generate_scenario_path(cmd_args.project_name, cmd_args.format)
    manager = ScenarioManager(
        pj,
        cmn,
        cmd_args.format,
        CommandArgument(args=cmd_args.execute_method_argument),
        project_name=cmd_args.project_name,
    )
    return manager.execute()



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from cliboa.adapter.postgres import PostgresqlAdaptor
from cliboa.scenario.rdbms import BaseRdbmsWrite


class PostgresqlWrite(BaseRdbmsWrite):
    def get_adaptor(self):
        return self._resolve(
            "adaptor_db",
            PostgresqlAdaptor,
            self.args.host,
            self.args.user,
            self.args.password,
            self.args.dbname,
            self.args.port,
        )



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from requests.auth import HTTPBasicAuth

from cliboa.scenario.base import BaseStep


class HttpBase(BaseStep):
    def __init__(self):
        super().__init__()
        self._src_url = None
        self._dest_dir = None
        self._dest_name = None
        self._timeout = 30
        self._retry_count = 3
        self._retry_intvl_sec = 10
        self._basic_auth = False
        self._user = None
        self._password = None

    def src_url(self, src_url):
        self._src_url = src_url

    def dest_dir(self, dest_dir):
        self._dest_dir = dest_dir

    def dest_name(self, dest_name):
        self._dest_name = dest_name

    def timeout(self, timeout):
        self._timeout = timeout

    def retry_count(self, retry_count):
        self._retry_count = retry_count

    def retry_intvl_sec(self, retry_intvl_sec):
        self._retry_intvl_sec = retry_intvl_sec

    def basic_auth(self, basic_auth):
        self._basic_auth = basic_auth

    def user(self, user):
        self._user = user

    def password(self, password):
        self._password = password

    def execute(self, *args):
        pass

    def get_params(self):
        if self._basic_auth:
            return {"auth": HTTPBasicAuth(self._user, self._password)}
        return {}



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import csv
import os

from pydantic import BaseModel

from cliboa.scenario.base import BaseStep
from cliboa.scenario.file import FileRead
from cliboa.util.exception import FileNotFound, InvalidParameter


class BaseRdbms(BaseStep):
    """
    Base class of relational database class.
    """

    class Arguments(BaseModel):
        host: str
        dbname: str
        user: str
        password: str
        port: int | None = None

    def execute(self):
        pass

    def get_adaptor(self):
        raise NotImplementedError("get_adaptor must be implemented in a subclass")

    def select_sql(self, tblname):
        """
        Create select sql.

        Arguments:
            tblname (str): table name

        Returns:
            str: SELECT * FROM {tblname}
        """
        return "SELECT * FROM %s" % tblname

    def insert_sql(self, tblname, columns):
        """
        Create insert sql from csv file.

        Arguments:
            tblname (str): table name
            columns (array): db column names

        Returns:
            str: INSERT INTO {tblname} (columns) VALUES (%s,%s,%s...)
        """
        item = ""
        for i, row in enumerate(columns):
            if i != 0:
                item = item + ","
            item = item + row
        place_holders = "%s," * (i + 1)
        return "INSERT INTO {} ({}) VALUES ({})".format(tblname, item, place_holders[:-1])

    def csv_as_params(self, path, chunk_size=100, encoding="utf-8"):
        """
        Divide csv records into small-blocked parameter.
        Basically it uses for paramer of insert query.

        Arguments:
            path (str): Csv file path
            chunk_size=100 (int): Number of row in a single block

        Returns:
            array: [
                    [(row1), (row2), (row3)]
                    [(row4), (row5), (row6)]
                    [(row7), (row8)]
                ]
        """
        tuples = []
        with open(path, mode="r", encoding=encoding) as f:
            reader = csv.reader(f)
            next(reader)  # ignore header
            rows = []
            for i, row in enumerate(reader, 1):
                rows.append(tuple(row))
                if i % chunk_size == 0:
                    tuples.append(rows)
                    rows = []
        if rows:
            tuples.append(rows)
        return tuples


class BaseRdbmsRead(BaseRdbms):
    class Arguments(BaseRdbms.Arguments):
        dest_path: str
        query: str | None = None
        tblname: str | None = None
        encoding: str = "UTF-8"

    def _property_path_reader(self, src, encoding="utf-8"):
        """
        Returns an resource contents from the path if src starts with "path:",
        returns src if not
        """
        self.logger.warning("DeprecationWarning: Will be removed in the near future")
        if src[:5].upper() == "PATH:":
            fpath = src[5:]
            if os.path.exists(fpath) is False:
                raise FileNotFound(src)
            with open(fpath, mode="r", encoding=encoding) as f:
                return f.read()
        return src

    def execute(self):
        if (self.args.query and self.args.tblname) or (
            not self.args.query and not self.args.tblname
        ):
            raise InvalidParameter("Either query or tblname is required.")

        dest_dir = os.path.dirname(self.args.dest_path)
        if dest_dir:
            os.makedirs(dest_dir, exist_ok=True)

        with self.get_adaptor() as adaptor:
            with open(self.args.dest_path, mode="w", encoding=self.args.encoding, newline="") as f:
                if self.args.tblname:
                    query = self.select_sql(self.args.tblname)
                elif isinstance(self.args.query, str):
                    self.logger.warning(
                        (
                            "DeprecationWarning: "
                            "In the near future, "
                            "the `query` will be changed to accept only dictionary types. "
                        )
                    )
                    query = self.args.property_path_reader(self.args.query)
                else:
                    query_filepath = self.args.source_path_reader(self.args.query)
                    with open(query_filepath, "r") as qf:
                        query = qf.read()
                cur = adaptor.select(query)
                writer = None
                for i, row in enumerate(cur):
                    if i == 0:
                        if type(row) is tuple:
                            writer = csv.writer(f, quoting=csv.QUOTE_ALL)
                            columns = [i[0] for i in cur.description]
                            writer.writerow(columns)
                        elif type(row) is dict:
                            writer = csv.DictWriter(f, list(row.keys()), quoting=csv.QUOTE_ALL)
                            writer.writeheader()
                    if writer:
                        writer.writerow(self.callback_handler(row))
                    else:
                        f.write(self.callback_handler(row))

    def callback_handler(self, row):
        if type(row) is tuple:
            return list(row)
        elif type(row) is dict:
            return row
        return row


class BaseRdbmsWrite(BaseRdbms, FileRead):
    class Arguments(BaseRdbms.Arguments, FileRead.Arguments):
        tblname: str
        encoding: str = "UTF-8"
        chunk_size: int = 100

    def execute(self):
        # Plural files are allowed to insert at the same time,
        # but all files must be the same csv format.
        files = self.get_src_files()
        if not self.check_file_existence(files):
            raise FileNotFound("No csv file was found.")

        with open(files[0], mode="r", encoding=self.args.encoding) as f:
            reader = csv.DictReader(f)
            fieldnames = reader.fieldnames

        query = self.insert_sql(self.args.tblname, fieldnames)
        self.logger.info("query: %s" % query)
        for file in files:
            with self.get_adaptor() as adaptor:
                tuples = self.csv_as_params(
                    file, chunk_size=self.args.chunk_size, encoding=self.args.encoding
                )
                self.logger.info("tuples: %s" % tuples)
                for params in tuples:
                    self.logger.info("params: %s" % params)
                    adaptor.insert(query, params)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os

from cliboa.adapter.http import Remove, Update, Upload
from cliboa.scenario.http import HttpBase
from cliboa.scenario.validator import EssentialParameters


class HttpPost(HttpBase):
    def __init__(self):
        super().__init__()
        self._payload = {}
        self._headers = {"accept": "application/json", "Content-Type": "application/json"}

    def payload(self, payload):
        self._payload = payload

    def headers(self, headers):
        self._headers = headers

    def execute(self, *args):
        os.makedirs(self._dest_dir, exist_ok=True)

        if self._basic_auth:
            valid = EssentialParameters(
                self.__class__.__name__,
                [self._src_url, self._dest_dir, self._dest_name, self._user, self._password],
            )
        else:
            valid = EssentialParameters(
                self.__class__.__name__, [self._src_url, self._dest_dir, self._dest_name]
            )
        valid()
        url = self._src_url
        dest_path = os.path.join(self._dest_dir, self._dest_name)

        u = Upload(
            url,
            dest_path,
            self._timeout,
            self._retry_count,
            self._retry_intvl_sec,
            params={
                "data": self._payload,
                "headers": self._headers,
            },
        )
        u.execute()


class HttpPut(HttpBase):
    def __init__(self):
        super().__init__()
        self._payload = {}
        self._headers = {"accept": "application/json", "Content-Type": "application/json"}

    def payload(self, payload):
        self._payload = payload

    def headers(self, headers):
        self._headers = headers

    def execute(self, *args):
        os.makedirs(self._dest_dir, exist_ok=True)

        if self._basic_auth:
            valid = EssentialParameters(
                self.__class__.__name__,
                [self._src_url, self._dest_dir, self._dest_name, self._user, self._password],
            )
        else:
            valid = EssentialParameters(
                self.__class__.__name__, [self._src_url, self._dest_dir, self._dest_name]
            )
        valid()
        url = self._src_url
        dest_path = os.path.join(self._dest_dir, self._dest_name)

        u = Update(
            url,
            dest_path,
            self._timeout,
            self._retry_count,
            self._retry_intvl_sec,
            params={
                "data": self._payload,
                "headers": self._headers,
            },
        )
        u.execute()


class HttpDelete(HttpBase):
    def __init__(self):
        super().__init__()

    def execute(self, *args):
        os.makedirs(self._dest_dir, exist_ok=True)

        if self._basic_auth:
            valid = EssentialParameters(
                self.__class__.__name__,
                [self._src_url, self._dest_dir, self._dest_name, self._user, self._password],
            )
        else:
            valid = EssentialParameters(
                self.__class__.__name__, [self._src_url, self._dest_dir, self._dest_name]
            )
        valid()
        url = self._src_url
        dest_path = os.path.join(self._dest_dir, self._dest_name)

        r = Remove(
            url,
            dest_path,
            self._timeout,
            self._retry_count,
            self._retry_intvl_sec,
            params=self.get_params(),
        )
        r.execute()



# flake8: noqa
#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from .base import BaseStepListener



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import json
from abc import ABC, abstractmethod
from time import sleep

import requests
from requests.exceptions import HTTPError

from cliboa.util.base import _BaseObject

VALID_HTTP_STATUS = 200


class Http(_BaseObject, ABC):
    """
    Http client abstract class
    """

    def __init__(self, url, dest_path, timeout, retry_cnt, retry_intvl_sec, params, **kwargs):
        super().__init__(**kwargs)
        self._url = url
        self._dest_path = dest_path
        self._timeout = timeout
        self._retry_cnt = retry_cnt
        self._retry_intvl_sec = retry_intvl_sec
        self._params = params

    @abstractmethod
    def request(self):
        pass

    def execute(self):
        self._logger.info("local path: %s" % self._dest_path)
        res = None
        for _ in range(self._retry_cnt):
            try:
                res = self.request()
                res.raise_for_status()
                with open(self._dest_path, "wb") as f:
                    f.write(res.content)
                return

            except Exception as e:
                self._logger.warning(e)

            self._logger.warning(
                "Unexpected error occurred during http request. Retry will start in %s",
                self._retry_intvl_sec,
            )
            sleep(self._retry_intvl_sec)

        if res is not None:
            raise HTTPError("Http request failed. HTTP Status code: %s" % res.status_code)
        else:
            raise HTTPError("Http request failed after all retries.")


class Download(Http):
    """
    Download via simple Http GET
    """

    def __init__(
        self,
        url,
        dest_path,
        timeout,
        retry_cnt=2,
        retry_intvl_sec=10,
        query_string=None,
        params={},
        **kwargs,
    ):
        super().__init__(url, dest_path, timeout, retry_cnt, retry_intvl_sec, params, **kwargs)
        # only when using request.get()
        self._query_string = query_string

    def request(self):
        self._logger.info("Http GET url: %s" % self._url)
        return requests.get(
            self._url, params=self._query_string, timeout=self._timeout, **self._params
        )


class Upload(Http):
    """
    Upload via simple Http POST
    """

    def __init__(
        self, url, dest_path, timeout, retry_cnt=2, retry_intvl_sec=10, params={}, **kwargs
    ):
        super().__init__(url, dest_path, timeout, retry_cnt, retry_intvl_sec, params, **kwargs)

    def request(self):
        self._logger.info("Http POST url: %s" % self._url)
        return requests.post(
            self._url,
            timeout=self._timeout,
            headers=self._params["headers"],
            data=json.dumps(self._params["data"]),
        )


class Update(Http):
    """
    Update via simple Http PUT
    """

    def __init__(
        self, url, dest_path, timeout, retry_cnt=2, retry_intvl_sec=10, params={}, **kwargs
    ):
        super().__init__(url, dest_path, timeout, retry_cnt, retry_intvl_sec, params, **kwargs)

    def request(self):
        self._logger.info("Http PUT url: %s" % self._url)
        return requests.put(
            self._url,
            timeout=self._timeout,
            headers=self._params["headers"],
            data=json.dumps(self._params["data"]),
        )


class Remove(Http):
    """
    Remove via simple Http DELETE
    """

    def __init__(
        self, url, dest_path, timeout, retry_cnt=2, retry_intvl_sec=10, params={}, **kwargs
    ):
        super().__init__(url, dest_path, timeout, retry_cnt, retry_intvl_sec, params, **kwargs)

    def request(self):
        self._logger.info("Http DELETE url: %s" % self._url)
        return requests.delete(self._url, timeout=self._timeout, **self._params)



import os

from cryptography.fernet import Fernet

from cliboa.adapter.file import File
from cliboa.scenario.transform.file import FileBaseTransform


class AesBase(FileBaseTransform):
    class Arguments(FileBaseTransform.Arguments):
        key_dir: str
        key_pattern: str

    def execute(self, *args):
        files = self.get_src_files()
        if not self.check_file_existence(files):
            self.logger.info("No files are found. Nothing to do.")
            return

        key_file = self._resolve("adapter_file", File).get_target_files(
            self.args.key_dir, self.args.key_pattern
        )
        if len(key_file) != 1:
            self.logger.info("Key file must be only one.")
            return

        with open("".join(key_file), "r") as k:
            key = k.read()

        for file in files:
            dest_path = os.path.join(self.args.resolve_dest_dir(), os.path.basename(file))
            self._process(file, dest_path, key)

    def _process(self, file, dest_path, key) -> None:
        raise NotImplementedError()


class AesEncrypt(AesBase):
    def _process(self, file, dest_path, key) -> None:
        with open(file, "r") as f:
            data = f.read()

        encrypt_data = Fernet(key.encode()).encrypt(data.encode())
        with open(dest_path, "wb") as w:
            w.write(encrypt_data)


class AesDecrypt(AesBase):
    def _process(self, file, dest_path, key) -> None:
        with open(file, "rb") as f:
            data = f.read()

        decrypt_data = Fernet(key.encode()).decrypt(data)

        with open(dest_path, "wb") as w:
            w.write(decrypt_data)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os
import re

from cliboa.adapter.azure import BlobServiceAdapter
from cliboa.scenario.azure import BaseAzureBlob
from cliboa.scenario.validator import EssentialParameters


class AzureBlobDownload(BaseAzureBlob):
    """
    Download from Azure blob storage
    """

    def __init__(self):
        super().__init__()
        self._prefix = ""
        self._src_pattern = None
        self._dest_dir = None

    def prefix(self, prefix):
        self._prefix = prefix

    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    def dest_dir(self, dest_dir):
        self._dest_dir = dest_dir

    def execute(self, *args):
        super().execute()

        valid = EssentialParameters(self.__class__.__name__, [self._src_pattern, self._dest_dir])
        valid()

        service = BlobServiceAdapter().get_client(
            account_url=self._account_url,
            account_access_key=self._account_access_key,
            connection_string=self._connection_string,
        )
        container_client = service.get_container_client(self._container_name)
        blobs = container_client.list_blobs(name_starts_with=self._prefix)
        for blob in blobs:
            filename = blob.name
            rec = re.compile(self._src_pattern)
            if not rec.fullmatch(filename):
                continue
            if self._dest_dir:
                os.makedirs(self._dest_dir, exist_ok=True)
            dest_path = os.path.join(self._dest_dir, os.path.basename(filename))
            blob_client = service.get_blob_client(container=self._container_name, blob=filename)

            with open(dest_path, "wb") as local_blob:
                blob_data = blob_client.download_blob()
                blob_data.readinto(local_blob)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os
import re

from cliboa.adapter.sftp import SftpAdapter
from cliboa.scenario.sftp import BaseSftp
from cliboa.util.base import _warn_deprecated, _warn_deprecated_args
from cliboa.util.constant import StepStatus


class SftpExtract(BaseSftp):
    def __init__(self):
        super().__init__()
        self.logger.warning(
            _warn_deprecated(
                "cliboa.scenario.load.extract.SftpExtract",
                "3.0",
                "4.0",
                "cliboa.scenario.sftp.BaseSftp",
            )
        )


class SftpDownload(BaseSftp):
    """
    Download files from sftp server
    """

    class Arguments(BaseSftp.Arguments):
        quit: bool = False
        ignore_empty_file: bool = False

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _quit(self):
        return self.args.quit

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _ignore_empty_file(self):
        return self.args.ignore_empty_file

    def execute(self, *args):
        os.makedirs(self.args.dest_dir, exist_ok=True)

        adapter = self.get_adapter()
        obj = adapter.list_files(
            dir=self.args.src_dir,
            dest=self.args.dest_dir,
            pattern=re.compile(self.args.src_pattern),
            endfile_suffix=self.args.endfile_suffix,
            ignore_empty_file=self.args.ignore_empty_file,
        )
        files = adapter.execute(obj)

        if self.args.quit is True and len(files) == 0:
            self.logger.info("No file was found. After process will not be processed")
            return StepStatus.SUCCESSFUL_TERMINATION

        self.logger.info("Files downloaded %s" % files)

        # cache downloaded file names
        self.put_to_context(files)


class SftpDelete(BaseSftp):
    """
    Delete file from sftp server
    """

    def execute(self, *args):
        adapter = self.get_adapter()
        obj = adapter.clear_files(
            dir=self.args.src_dir,
            pattern=re.compile(self.args.src_pattern),
        )
        adapter.execute(obj)


class SftpDownloadFileDelete(BaseSftp):
    """
    Delete all downloaded files.
    """

    Arguments = None

    def execute(self, *args):
        files = self.get_from_context()
        if files is None or len(files) == 0:
            self.logger.info("No files to delete.")
            return

        self.logger.info("Delete files %s" % files)
        adapter = self._resolve(
            "adapter_sftp",
            SftpAdapter,
            host=self.get_symbol_argument("host"),
            user=self.get_symbol_argument("user"),
            password=self.get_symbol_argument("password"),
            key=self.get_symbol_argument("key"),
            passphrase=self.get_symbol_argument("passphrase"),
            timeout=self.get_symbol_argument("timeout"),
            retryTimes=self.get_symbol_argument("retry_count"),
            port=self.get_symbol_argument("port"),
        )
        symbol_src_dir = self.get_symbol_argument("src_dir")
        symbol_endfile_suffix = self.get_symbol_argument("endfile_suffix")

        for file in files:
            obj = adapter.remove_specific_file(
                dir=symbol_src_dir,
                fname=file,
            )
            adapter.execute(obj)
            self.logger.info("%s is successfully deleted." % file)

            if symbol_endfile_suffix:
                obj = adapter.remove_specific_file(
                    dir=symbol_src_dir,
                    fname=file + symbol_endfile_suffix,
                )
                self.logger.info("%s is successfully deleted." % (file + symbol_endfile_suffix))


class SftpFileExistsCheck(BaseSftp):
    """
    File check in sftp server
    """

    class Arguments(BaseSftp.Arguments):
        ignore_empty_file: bool = False

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _ignore_empty_file(self):
        return self.args.ignore_empty_file

    def execute(self, *args):
        adapter = self.get_adapter()
        obj = adapter.file_exists_check(
            dir=self.args.src_dir,
            pattern=re.compile(self.args.src_pattern),
            ignore_empty_file=self.args.ignore_empty_file,
        )
        files = adapter.execute(obj)

        if len(files) == 0:
            self.logger.info("File not found. After process will not be processed")
            return StepStatus.SUCCESSFUL_TERMINATION

        self.logger.info("File was found. After process will be processed")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#

import argparse
import os
from pathlib import Path
from shutil import copyfile

import cliboa


class CliboAdmin:
    """
    cliboa administrator command
    """

    def __init__(self, args):
        self._args = args
        self._cwd = Path.cwd().resolve()
        print(f"Working root dir is {self._cwd}")
        self._cliboa_install_path = os.path.dirname(cliboa.__path__[0])

    def main(self):
        """
        Usage:
            cliboadmin init $app_name
            cliboadmin project|pj $project_name
        """
        if self._args.cmd == "init":
            self._init_project(self._args.app)
        elif self._args.cmd in ("project", "pj"):
            if self._args.project:
                self._create_new_project(self._args.project)
            else:
                print("Please specify -p PROJECT_NAME.")

    def _init_project(self, app_name: str | None):
        """
        Initialize program configuration
        """
        self._create_dir("common")
        self._create_dir("project")
        if self._create_dir("conf"):
            self._copy_file(
                self._get_cliboa_path("cli", "template", "logging.conf"),
                self._get_dest_path("conf", "logging.conf"),
            )
            self._create_dir("logs")
        if app_name:
            self._create_dir(app_name)
            dest_path = self._get_dest_path(app_name, "cliboa_run.py")
            self._copy_file(self._get_cliboa_path("cli", "template", "cliboa_run.py"), dest_path)
            self._replace_app_in_file(dest_path, app_name)
            self._copy_file(
                self._get_cliboa_path("conf", "default_environment.py"),
                self._get_dest_path(app_name, "cliboa_environment.py"),
            )

    def _create_dir(self, *args) -> bool:
        target = self._get_dest_path(*args)
        print(f"Create {target}")
        try:
            if os.path.isdir(target):
                print(f"Already exists {target}")
                return True
            os.makedirs(target, exist_ok=False)
            return True
        except Exception as e:
            print(f"Warning: Failed create dir {target}: {e}")
            return False

    def _copy_file(self, src: str, dest: str) -> bool:
        try:
            print(f"Create {dest}")
            if os.path.isfile(dest):
                print(f"Warning: Already exists {dest}")
                return False
            copyfile(src, dest)
            return True
        except Exception as e:
            print(f"Warning: Failed create file {dest}: {e}")
            return False

    def _get_cliboa_path(self, *args):
        return os.path.join(self._cliboa_install_path, "cliboa", *args)

    def _get_dest_path(self, *args):
        return os.path.join(self._cwd, *args)

    def _replace_app_in_file(self, dest_path: str, app_name: str) -> bool:
        """
        Convert "app.cliboa_environment"
        """
        if not os.path.exists(dest_path):
            return False
        original_string = "app.cliboa_environment"
        replacement_string = f"{app_name}.cliboa_environment"
        try:
            with open(dest_path, "r", encoding="utf-8") as f:
                content = f.read()
            new_content = content.replace(original_string, replacement_string)
            if new_content != content:
                with open(dest_path, "w", encoding="utf-8") as f:
                    f.write(new_content)
            return True
        except Exception:
            print(f"Warning: Failed update {dest_path} for {app_name}")
            return False

    def _create_new_project(self, pj_name: str):
        """
        Create an individual project configuration
        """
        if self._create_dir("project", pj_name):
            self._copy_file(
                self._get_cliboa_path("cli", "template", "scenario-sample.yml"),
                self._get_dest_path("project", pj_name, "scenario.yml"),
            )
        self._create_dir("project", pj_name, "scenario")


class CommandArgumentParser:
    def parse(self):
        """
        Parse cliboadmin arguments
        """
        parser = argparse.ArgumentParser()
        parser.add_argument("cmd", choices=["init", "project", "pj"], help="init or project(pj)")
        parser.add_argument("-a", "--app", help="Your application name")
        parser.add_argument("-p", "--project", help="Project name")
        return parser.parse_args()


def main():
    parser = CommandArgumentParser()
    args = parser.parse()
    admin = CliboAdmin(args)
    admin.main()


if __name__ == "__main__":
    main()



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from typing import Any

from cliboa.core.interface import _IContext
from cliboa.util.base import _BaseObject


class _CliboaContext(_BaseObject, _IContext):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._data = {}

    def put(self, key: str, value: Any) -> None:
        self._data[key] = value

    def get(self, key: str) -> Any:
        return self._data.get(key)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from cliboa import state
from cliboa.listener.base import BaseScenarioListener
from cliboa.listener.interface import IScenarioExecutor


class ScenarioStatusListener(BaseScenarioListener):
    """
    Listener for scenario execution status
    """

    def before(self, executor: IScenarioExecutor) -> None:
        state.set("_ExecuteScenario")
        self.logger.info(f"Start scenario execution. StepQueue size is {executor.max_steps_size}")

    def after(self, executor: IScenarioExecutor) -> None:
        state.set("_ExecuteScenario")

    def error(self, executor: IScenarioExecutor, e: Exception) -> None:
        state.set("_ExecuteScenario")

    def completion(self, executor: IScenarioExecutor) -> None:
        state.set("_ExecuteScenario")
        self.logger.info(
            f"Complete scenario execution. StepQueue size is {executor.current_steps_size}"
        )



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os

##################################################
# 1. File/Directory - used within cliboa.interface.run
##################################################

# INTERNAL: app base dir. You can remove BASE_DIR if you want.
BASE_DIR = os.getcwd()

# OPTIONAL: scenario file name excluding extension
SCENARIO_FILE_NAME = "scenario"

# OPTIONAL: scenario yaml file ext
SCENARIO_YAML_EXT = ".yml"

# OPTIONAL: Project directory path.
PROJECT_DIR = os.path.join(BASE_DIR, "project")

# Common scenario directory path.
COMMON_DIR = os.path.join(BASE_DIR, "common")

##################################################
# 2. Logging
##################################################

# OPTIONAL: logging config file used for logging.config.fileConfig within cliboa.interface.run.
# This setting takes precedence over {BASE_DIR}/conf/logging.conf if defined.
LOGGING_CONFIG_PATH = os.path.join(BASE_DIR, "conf", "logging.conf")

# OPTIONAL: logging config dict used for logging.config.dictConfig within cliboa.interface.run.
# This setting takes precedence over LOGGING_CONFIG_PATH if defined.
# LOGGING_CONFIG_DICT = {}

# OPTIONAL: logging mask re pattern.
# If not defined (commented out), the default pattern is applied.
# LOGGING_MASK = ".*password.*|.*secret_key.*"

# OPTIONAL: logging partial mask re pattern.
# If not defined (commented out), the default pattern is applied.
# LOGGING_PARTIAL_MASK = ".*access_key.*|.*token.*"

# OPTIONAL: logging partial mask config.
# expose the first and last LOGGING_PARTIAL_NUM characters.
# If not defined (commented out), the default value is applied.
# LOGGING_PARTIAL_NUM = 3

##################################################
# 3. Scenario class - used for import custom classes
##################################################

# REQUIRED: common custom classes's root module paths.
# remove "cliboa.scenario" and add your custom classes's module root paths.
COMMON_CUSTOM_ROOT_PATHS = ["cliboa.scenario"]

# REQUIRED: common custom classes's root module paths.
PROJECT_CUSTOM_ROOT_PATHS = ["project"]

# REQUIRED: Directory name which scenario python file is placed,
# such as f"{project_custom_root_path}/{project_name}/{SCENARIO_DIR_NAME}/project_step.py".
# If commented out, f"project/{project_name}/project_step.py"
# v2 backward compatibility: Alternate use SCENARIO_DIR_NAME
#                            if PROJECT_SCENARIO_DIR_NAME is not defined.
PROJECT_SCENARIO_DIR_NAME = "scenario"

# REQUIRED: common custom classes to make available.
# You can define empty list this value if you do not use common custom classes.
COMMON_CUSTOM_CLASSES = ["sample_step.SampleStep", "sample_step.SampleStepSub"]

# REQUIRED: project congenital classes to make available.
# You can define empty list this value if you do not use project custom classes.
PROJECT_CUSTOM_CLASSES = []



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os
import re
import subprocess
from typing import Any, Tuple

from pydantic import BaseModel, Field, PrivateAttr, field_validator, model_validator

# 'typing.Self' is available in Python 3.11+
from typing_extensions import Self

from cliboa.util.base import _warn_deprecated
from cliboa.util.exception import InvalidFormat, InvalidParameter, ScenarioFileInvalid


class _BaseWithVars(BaseModel):
    with_vars: dict[str, str] = Field(default_factory=dict)
    _with_static_vars: dict[str, str] = PrivateAttr(default_factory=dict)

    def calc(self) -> None:
        """
        calculate with_vars shell commands, and store result.
        """
        for var_name, cmd in self.with_vars.items():
            self._with_static_vars[var_name] = self._exec_shell_cmd(cmd)

    def _exec_shell_cmd(self, cmd: str):
        """
        exec a shell script and return result string to replace as variable.
        """
        shell_output = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, shell=True
        ).communicate()[  # nosec
            0
        ]
        shell_output = shell_output.strip()
        # remove head byte string
        shell_output = re.sub("^b", "", str(shell_output))
        # remove '
        shell_output = re.sub("'", "", str(shell_output))
        return shell_output

    def _merge_static_vars(self, data: dict[str, str]) -> None:
        """
        merge calc result (not overwrite)
        """
        self._with_static_vars = data | self._with_static_vars


_warned_with_vars: bool = False


class StepModel(_BaseWithVars):
    step: str = Field(frozen=True)
    class_name: str = Field(alias="class", frozen=True)
    listeners: str | list[str] | None = None
    symbol: str | None = Field(default=None, frozen=True)
    arguments: dict[str, Any] = Field(default_factory=dict)

    @model_validator(mode="before")
    @classmethod
    def _extract_with_vars(cls, data: Any) -> Any:
        """v2 compatibility"""
        if not isinstance(data, dict):
            return data
        arguments = data.get("arguments")
        if isinstance(arguments, dict) and "with_vars" in arguments:
            if "with_vars" in data:
                raise InvalidFormat("duplicate definition 'with_vars'")
            data["with_vars"] = arguments.pop("with_vars")
            global _warned_with_vars
            if not _warned_with_vars:
                _warn_deprecated(
                    "scenario file's scenario.[].arguments.with_vars",
                    "3.0",
                    "4.0",
                    "scenario.[].with_vars",
                )
                _warned_with_vars = True
        return data

    def get_listeners(self) -> list[str]:
        """
        Get the listeners in list format.
        """
        if isinstance(self.listeners, str):
            return [self.listeners]
        elif isinstance(self.listeners, list):
            return self.listeners
        else:
            return []

    def replace_vars(self) -> None:
        """
        Replace variable expressions in arguments.
        This is need to be called after call calc.
        """
        self.arguments = self._replace_arguments(self.arguments)

    def _replace_arguments(self, arguments: Any) -> Any:
        """
        Replace nested argments.
        """
        if isinstance(arguments, dict):
            return {k: self._replace_arguments(v) for k, v in arguments.items()}
        elif isinstance(arguments, list):
            return [self._replace_arguments(v) for v in arguments]
        elif isinstance(arguments, str):
            matches = re.compile(r"{{(.*?)}}").findall(arguments)
            for match in matches:
                var_name = match.strip()
                arguments = self._replace_vars(arguments, var_name)
            return arguments
        else:
            return arguments

    def _replace_vars(self, value: str, var_name: str):
        """
        This method replaces the value of {{ xxx }}
        """
        if not var_name:
            raise InvalidParameter("name in variable expression was empty.")
        if var_name.startswith("env."):
            replace_str = os.environ.get(var_name[4:])
        else:
            replace_str = self._with_static_vars.get(var_name)
        if replace_str is None:
            raise ScenarioFileInvalid(
                "scenario file is invalid." " variable '%s' can not be resolved." % var_name
            )
        return re.sub(r"{{(\s?)%s(\s?)}}" % var_name, replace_str, value)


class ParallelConfigModel(BaseModel):
    multi_process_count: int | None = Field(default=None, ge=2)
    force_continue: bool | None = None

    def merge(self, model: Self) -> None:
        """
        Merge model's props (only when self value is None)
        """
        if not isinstance(model, ParallelConfigModel):
            return
        for k, v in self.model_dump().items():
            if v is None:
                r = getattr(model, k)
                if r is not None:
                    setattr(self, k, r)

    def fill_default(self) -> Self:
        """
        Set defalut values if they are None
        """
        if self.multi_process_count is None:
            self.multi_process_count = 2
        if self.force_continue is None:
            self.force_continue = False
        return self


class ParallelStepModel(BaseModel):
    step: str | None = Field(default=None, frozen=True)
    parallel: Tuple[StepModel, ...] = Field(min_length=1, frozen=True)
    parallel_config: ParallelConfigModel = Field(default_factory=ParallelConfigModel)

    def _merge_parallel_config(self, data: ParallelConfigModel) -> None:
        """
        merge ParallelConfig to under this model.
        """
        if not isinstance(data, ParallelConfigModel):
            return
        self.parallel_config.merge(data)


class ScenarioModel(_BaseWithVars):
    scenario: list[StepModel | ParallelStepModel, ...]
    parallel_config: ParallelConfigModel = Field(default_factory=ParallelConfigModel)

    def merge(self, cmn: Self) -> None:
        """
        Merge common scenario settings.
        """
        self.parallel_config.merge(cmn.parallel_config)
        self.with_vars = cmn.with_vars | self.with_vars

        for step in self.scenario:
            if isinstance(step, StepModel):
                self._merge_step(step, cmn)
            elif isinstance(step, ParallelStepModel):
                for p_step in step.parallel:
                    self._merge_step(p_step, cmn)

    def _merge_step(self, step: StepModel, cmn: Self) -> None:
        """shallow marge"""
        for cmn_step in cmn.scenario:
            if not isinstance(cmn_step, StepModel):
                continue
            if step.class_name != cmn_step.class_name:
                continue

            step.arguments = cmn_step.arguments | step.arguments
            step.with_vars = cmn_step.with_vars | step.with_vars
            return

    def setup(self) -> None:
        """
        Prepare to use scenario.

        1. calc scenario's with_vars
        2. calc step's  with_vars
        3. propagate calculated with_vars and parallel_config from scenario to steps
        4. replace step's arguments with calculated with_vars
        """
        self.calc()
        self._calc_steps()
        self._propagate()
        self._replace_vars_steps()

    def _apply_steps(self, func: str, *args, **kwargs) -> None:
        for step in self.scenario:
            if isinstance(step, StepModel):
                getattr(step, func)(*args, **kwargs)
            elif isinstance(step, ParallelStepModel):
                for p_step in step.parallel:
                    getattr(p_step, func)(*args, **kwargs)

    def _apply_parallel_steps(self, func: str, *args, **kwargs) -> None:
        for step in self.scenario:
            if isinstance(step, ParallelStepModel):
                getattr(step, func)(*args, **kwargs)

    def _calc_steps(self) -> None:
        """
        exec 'calc' method in all steps.
        """
        self._apply_steps("calc")

    def _propagate(self) -> None:
        """
        merge _with_static_vars and parallel_config to scenario steps.
        """
        if self._with_static_vars:
            self._apply_steps("_merge_static_vars", self._with_static_vars)
        self._apply_parallel_steps("_merge_parallel_config", self.parallel_config)

    def _replace_vars_steps(self) -> None:
        """
        exec 'replace_vars' method in all steps.
        """
        self._apply_steps("replace_vars")


class CommandArgument(BaseModel):
    args: list[Any] = Field(default_factory=list)
    kwargs: dict[str, Any] = Field(default_factory=dict)

    @field_validator("args", mode="before")
    @classmethod
    def convert_single_string_to_list(cls, value):
        if not isinstance(value, list) and isinstance(value, str):
            if value == "":
                return []
            return [value]
        return value



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from pydantic import BaseModel

from cliboa.scenario.base import BaseStep
from cliboa.util.base import _warn_deprecated_args


class BaseGcp(BaseStep):
    """
    Base class of Gcp usage.
    """

    class Arguments(BaseModel):
        project_id: str
        credentials: str

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _project_id(self):
        return self.args.project_id

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _credentials(self):
        return self.args.credentials

    def get_credentials(self):
        return self.args.credentials


class BaseBigQuery(BaseGcp):
    """
    Base class of BigQuery use.

    """

    class Arguments(BaseGcp.Arguments):
        dataset: str
        tblname: str
        location: str

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _dataset(self):
        return self.args.dataset

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _tblname(self):
        return self.args.tblname

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _location(self):
        return self.args.location


class BaseGcs(BaseGcp):
    """
    Base class of Gcs use.

    """

    class Arguments(BaseGcp.Arguments):
        bucket: str

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _bucket(self):
        return self.args.bucket


class BaseFirestore(BaseGcp):
    """
    Base class of Firebase use.

    """

    class Arguments(BaseGcp.Arguments):
        collection: str
        document: str | None = None
